from .tools.typing.response import ForeFrontResponse
from .attributes.conversation import Conversation
from .tools.system.signature import Encrypt
from typing import Optional, Union, Generator, Dict, List
from base64 import b64encode
import fake_useragent
import tls_client
import requests
import uuid
import json
import logging

class Model:
	@classmethod
	def __init__(self: object, sessionID: str, client: str, model: Optional[str] = "gpt-3.5-turbo", 
		conversationID: Optional[Union[str, None]] = None
	) -> None:
		self._SETUP_LOGGER()
		self.Conversation: Conversation = Conversation(model=self)
		self._session: requests.Session = requests.Session()
		self._model: str = model
		self._API = "https://chat-api.tenant-forefront-default.knative.chi.coreweave.com/api/trpc"
		self.__NAME: Union[str, None] = None
		self._WORKSPACEID: str = ''
		self._USERID: str = "user_"
		self._CLIENT: str = client
		self._SESSION_ID: str = sessionID
		self.CONVERSATION_ID: [Union[str, None]] = conversationID
		self._PERSONA: str = "607e41fe-95be-497e-8e97-010a59b2e2c0"
		self._JSON: Dict[str, str] = {}
		self._HEADERS: Dict[str, str] = {
			"Authority": "streaming.tenant-forefront-default.knative.chi.coreweave.com",
			"Accept": "*/*",
			"Accept-Language": "en,pt-BR,fr-FR;q=0.9,fr;q=0.8,es-ES;q=0.7,es;q=0.6,en-US;q=0.5,am;q=0.4,de;q=0.3",
			"Authorization": f"Bearer {self._CLIENT}",
			"Cache-Control": "no-cache",
			"Pragma": "no-cache",
			"Content-Type": "application/json",
			"Origin": "https://chat.forefront.ai",
			"Referer": "https://chat.forefront.ai/",
			"Sec-Ch-Ua": "\"Chromium\";v=\"112\", \"Not:A-Brand\";v=\"99\"",
			"Sec-Ch-Ua-mobile": "?0",
			"Sec-Ch-Ua-platform": "\"macOS\"",
			"Sec-Fetch-Dest": "empty",
			"Sec-Fetch-Mode": "cors",
			"Sec-Fetch-Site": "cross-site",
			"User-Agent": fake_useragent.UserAgent().random
		}

		self._JWT_HEADERS: Dict[str, str] = {
			"Authority": "clerk.forefront.ai",
			"Accept": "*/*",
			"Cache-Control": "no-cache",
			"Content-Type": "application/x-www-form-urlencoded",
			"Origin": "https://chat.forefront.ai",
			"Pragma": "no-cache",
			"Cookie": f"__client={self._CLIENT}",
			"Sec-Ch-Ua": "\"Chromium\";v=\"112\", \"Google Chrome\";v=\"112\", \"Not:A-Brand\";v=\"99\"",
			"Sec-Ch-Ua-mobile": "?0",
			"Sec-Ch-Ua-platform": "\"macOS\"",
			"Referer": "https://chat.forefront.ai/",
			"Sec-Fetch-Dest": "empty",
			"Sec-Fetch-Mode": "cors",
			"Sec-Fetch-Site": "same-site",
			"User-Agent": fake_useragent.UserAgent().random
		}

		self._WORKSPACEID = self._GetWorkspaceID()
		self._USERID = self._GetUserID()
		self._logger.debug("Connected in Workspace: " + self._WORKSPACEID)

	@classmethod
	def _SETUP_LOGGER(self: type) -> None:
		self._logger: logging.getLogger = logging.getLogger(__name__)
		self._logger.setLevel(logging.DEBUG)
		console_handler: logging.StreamHandler = logging.StreamHandler()
		console_handler.setLevel(logging.DEBUG)
		formatter: logging.Formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		console_handler.setFormatter(formatter)

		self._logger.addHandler(console_handler)

	@classmethod
	def _UpdateJWTToken(self: type) -> None:
		jwt_token: Dict[str, str] = {}
		jwt_status: int = 0

		while True:
			jwt_token = self._session.post(f"https://clerk.forefront.ai/v1/client/sessions/{self._SESSION_ID}/tokens?_clerk_js_version=4.38.4", 
			headers=self._JWT_HEADERS)
			jwt_status = jwt_token.status_code

			if jwt_status == 200:
				break

		self._HEADERS["Authorization"] = f"Bearer {jwt_token.json()['jwt']}"

	@classmethod
	def _UpdateXSignature(self: type) -> None:
		DATA_: str = b64encode((self._USERID + self._PERSONA + self._WORKSPACEID).encode()).decode()
		self._HEADERS["X-Signature"] = Encrypt(DATA_, self._SESSION_ID)

	@classmethod
	def _GetUserID(self: type) -> str:
		DATA_: Dict[str, str] = self._session.post(f"https://clerk.forefront.ai/v1/client/sessions/{self._SESSION_ID}/touch?_clerk_js_version=4.38.4",
																headers=self._JWT_HEADERS).json()
		return DATA_["response"]["user"]["id"]

	@classmethod
	def _GetWorkspaceID(self: type) -> str:
		self._UpdateJWTToken()
		url: str = f"{self._API}/workspaces.listWorkspaces,chat.loadTree?batch=1&input="
		payload: str = "{\"0\":{\"json\":null,\"meta\":{\"values\":[\"undefined\"]}},\"1\":{\"json\":{\"workspaceId\":\"\"}}}"
		return self._session.get(url + payload, headers=self._HEADERS).json()[0]["result"]["data"]["json"][0]["id"]

	@classmethod
	def SetupConversation(self: type, prompt: str, options: Optional[Dict[str, str]] = {}) -> None:
		action = "new"
		conversations: Dict[str, str] = self.Conversation.GetList()
		if self.CONVERSATION_ID is None:
			if conversations[-1]["type"] == "chat":
				self.CONVERSATION_ID = conversations[-1]["id"]
				action = "continue"
		else:
			action = "continue"

		if "create" in options:
			if options["create"]:
				action = "new"

				if "name" not in options:
					self.__logger.error("Invalid options.")
					return None

				for cv in conversations:
					if cv["name"].lower() != options["name"].lower():
						self.__NAME = options["name"]

		self._JSON = {
			"text": prompt,
			"action": action,
			"id": self.CONVERSATION_ID,
			"parentId": self._WORKSPACEID,
			"workspaceId": self._WORKSPACEID,
			"messagePersona": self._PERSONA,
			"model": self._model
		}

	@classmethod
	def IsAccountActive(self: type) -> bool:
		return self._session.post(f"https://clerk.forefront.ai/v1/client/sessions/{self._SESSION_ID}/touch?_clerk_js_version=4.38.4", 
			headers=self._JWT_HEADERS).status_code == 200

	@classmethod
	def SendConversation(self: type) -> Generator[ForeFrontResponse, None, None]:
		self._UpdateJWTToken()
		self._UpdateXSignature()

		for chunk in self._session.post("https://streaming.tenant-forefront-default.knative.chi.coreweave.com/chat", 
			headers=self._HEADERS, json=self._JSON, stream=True
		).iter_lines():
			if b"finish_reason\":null" in chunk:
				data = json.loads(chunk.decode('utf-8').split("data: ")[1])
				yield ForeFrontResponse(**data)

		conversations: List[Dict[str, str]] = self.Conversation.GetList()
		if self.__NAME is not None:
			self.Conversation.Rename(conversations[-1]["id"], self.__NAME)
			self.__NAME = None
		else:
			if conversations[-1]["name"].lower() == "new chat":
				conversation: Dict[str, str] = conversations[-1]
				self.Conversation.Rename(conversation["id"], self.Conversation.GenerateName(self._JSON["text"]))


from typing import List, Optional, Dict, Union
from .tools.typing.response import ModelResponse
import sys, re
import uuid
import logging
import tls_client
import fake_useragent
import json

class Model:
	@classmethod
	def __init__(self: type) -> None:
		self.__SETUP_LOGGER()
		self.__response: Dict[str, str] = {"prompt": '', "answer": '', "externalLinks": []}
		self.__session: object = tls_client.Session(client_identifier="chrome_108")
		self.__URI: str = "https://you.com/api/streamingSearch"
		self.__USER_AGENT: str = fake_useragent.UserAgent().random
		self.__ID: str = str(uuid.uuid4())
		self.__PARAMS: Dict[str, str] = {"mkt": ''}
		self.__HEADERS: Dict[str, str] = {
			"Access-Control-Allow-Origin": "*",
			"Authority": "you.com",
			"Accept": "text/event-stream",
			"Accept-Language": "pt-BR,en,fr-FR;q=0.9,fr;q=0.8,es-ES;q=0.7,es;q=0.6,en-US;q=0.5,am;q=0.4,de;q=0.3",
			"Accept-Encoding": "gzip, deflate, br",
			"Cache-Control": "no-cache, no-transform",
			"CF-Cache-Status": "DYNAMIC",
			"Referer": "https://you.com/search?q=who+are+you&tbm=youchat&cfr=chat",
			"Sec-Ch-Ua": "\"Chromium\";v=\"112\", \"Not:A-Brand\";v=\"99\"",
			"Sec-Ch-Ua-Mobile": "?0",
			"Sec-Ch-Ua-Plataform": "\"Windows\"",
			"Sec-Fetch-Dest": "empty",
			"Sec-Fetch-Mode": "cors",
			"Sec-Fetch-Site": "same-origin",
			"Alt-Svc": "h3=\":443\"; ma=86400, h3-29=\":443\"; ma=86400",
			"User-Agent": self.__USER_AGENT,
			"Cookie": f"safesearch_guest=Moderate; uuid_guest={self.__ID}"		
		}

	@classmethod
	def __SETUP_LOGGER(self: type) -> None:
		self.__logger: logging.getLogger = logging.getLogger(__name__)
		self.__logger.setLevel(logging.DEBUG)
		console_handler: logging.StreamHandler = logging.StreamHandler()
		console_handler.setLevel(logging.DEBUG)


		formatter: logging.Formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		console_handler.setFormatter(formatter)

		self.__logger.addHandler(console_handler)

	@classmethod
	def SetupConversation(self: type, prompt: str, history: Optional[List[dict]] = []) -> None:
		self.__PARAMS: Dict[str, Union[str, int, bool]] = {
			'q': prompt,
			"page": 1,
			"count": 10,
			"safeSearch": "Moderate",
			"onShoppingPage": False,
			"mkt": '',
			"responseFilter": "WebPages,Translations,TimeZone,Computation,RelatedSearches",
			"domain": "youchat",
			"queryTraceId": self.__ID,
			"chat": str(history),
			"chatId": self.__ID,
		}

		self.__response["prompt"] = prompt

	@classmethod
	def __ERROR_ON_RESPONSE(self: type) -> None:
		self.__logger.error("Unable to fetch the response, Please try again.")
		sys.exit(1)

	@classmethod
	def SendConversation(self: type) -> ModelResponse:
		self.__session.headers = self.__HEADERS

		output: object = self.__session.get(self.__URI, params=self.__PARAMS)

		if "youChatToken" not in output.text:
			self.__ERROR_ON_RESPONSE()

		idx: int = 0
		events: List[str] = output.text.split('\n')

		text: str= ''
		externalLinks: List[Dict[str, str]] = json.loads(re.search(
			r"(?<=event: thirdPartySearchResults\ndata:)(.*\n)*?(?=event: )", output.text
			).group())["search"]["third_party_search_results"]

		while events[idx].strip() != "event: done":
		   if events[idx].strip() == "event: youChatToken":
		       idx += 1
		       if events[idx].startswith("data:"):
		           text += json.loads(events[idx].replace("data:", '').strip())["youChatToken"]
		   idx += 1

		self.__response["answer"] = text
		self.__response["externalLinks"] = externalLinks
		
		return ModelResponse(**self.__response)



import requests, time, ast, json
from bs4 import BeautifulSoup
from hashlib import sha256

class Model:
    # answer is returned with html formatting
    next_id = None
    messages = []
    answer = None
    
    def __init__(self):
        r = requests.get("https://italygpt.it")
        soup = BeautifulSoup(r.text, "html.parser")
        self.next_id = soup.find("input", {"name": "next_id"})["value"]

    def GetAnswer(self, prompt: str, messages: list = []):
        try:
            r = requests.get("https://italygpt.it/question", params={"hash": sha256(self.next_id.encode()).hexdigest(), "prompt": prompt, "raw_messages": json.dumps(messages)}).json()
        except Exception as e:
            r = requests.get("https://italygpt.it/question", params={"hash": sha256(self.next_id.encode()).hexdigest(), "prompt": prompt, "raw_messages": json.dumps(messages)}).text
            if "too many requests" in r.lower():
                # rate limit is 5 requests per 1 minute
                time.sleep(20)
                return self.create(prompt, messages)
            else:
                print(f"There was an error getting the response, consider re-initializing the italygpt instance. {e}")
                return
        if r.get("error") != None:
            error = r["error"]
            if "ip is banned" in error.lower():
                print("Your ip was banned. Support email is: support@ItalyGPT.it")
                return
            elif "proxy not allowed" in error.lower():
                print("You are using a proxy, italygpt api calls do not allow proxies.")
                return
            elif "prompt too long" in error.lower():
                print("Your prompt is too long (max characters is: 1000)")
                return
            else:
                print(f"There was an error with your request: {error}")
                return
        self.next_id = r["next_id"]
        self.messages = ast.literal_eval(r["raw_messages"])
        self.answer = r["response"]
        return self


from pydantic import BaseModel
from typing import Optional, Union, Dict, List

class ExternalLinks(BaseModel):
	name: str
	url: str
	displayUrl: str
	snippet: str
	language: Optional[Union[str, None]] = ''
	thumbnailUrl: Optional[Union[str, None]] = ''
	isFamilyFriendly: Optional[Union[str, None]] = ''
	isNavigational: Optional[Union[str, None]] = ''
	snmixLink: Optional[Union[str, None]] = ''

class ModelResponse(BaseModel):
	prompt: str
	answer: Optional[str] = ''
	externalLinks: Optional[List[ExternalLinks]] = []